# pylint: disable=missing-docstring,line-too-long,no-self-use
"""proxy_test.py"""

import unittest
import itertools
from nose.tools import assert_raises # pylint: disable=no-name-in-module

from .proxy import Proxy
from . import rpc
from . import exceptions

class ProxyTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods

    def setUp(self):
        rpc.id_generator = itertools.count(1) # First generated is 1

    def test_request(self):
        proxy = Proxy('http://non-existant/')
        with assert_raises(exceptions.ConnectionError):
            proxy.request('add', 1, 2)

    def test_notify(self):
        proxy = Proxy('http://non-existant/')
        with assert_raises(exceptions.ConnectionError):
            proxy.notify('add', 1, 2)

    def test_magic(self):
        proxy = Proxy('http://non-existant/')
        with assert_raises(exceptions.ConnectionError):
            proxy.add(1, 2)

    def test_handle_response_none(self):
        with assert_raises(exceptions.ParseError):
            Proxy.handle_response('{')



# pylint: disable=missing-docstring,line-too-long
"""rpc_test.py

This should test:

Notifications
    rpc.request('go')
    rpc.request('print', 'Hello world')
    rpc.request('print', 'Hello', 'world')
    rpc.request('print', string='Hello world')
    rpc.request('print', string1='Hello', string2='world')
    rpc.request('set_age', 40, name='Beau Barker')

Requests (requiring a response):
    rpc.request('get', response=True)
    rpc.request('sqrt', 1, response=True)
    rpc.request('add', 1, 2, response=True)
    rpc.request('find', name='Beau', response=True)
    rpc.request('find', name='Beau', age=38, response=True)
    rpc.request('set_age', 40, name="Beau Barker", response=True)
"""

import unittest
import itertools
from nose.tools import assert_equal # pylint: disable=no-name-in-module

from . import rpc

class RPCTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods,no-self-use

    def setUp(self):
        rpc.id_generator = itertools.count(1) # First generated is 1

    # Notifications

    def test_method_only(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "go"},
            rpc.request('go')
        )

    def test_one_arg(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "print", "params": ["Hello world"]},
            rpc.request('print', 'Hello world')
        )

    def test_two_args(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "print", "params": ["Hello", "world"]},
            rpc.request('print', 'Hello', 'world')
        )

    def test_one_kwarg(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "print", "params": {"string": "Hello world"}},
            rpc.request('print', string='Hello world')
        )

    def test_two_kwargs(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "print", "params": {"string1": "Hello", "string2": "world"}},
            rpc.request('print', string1='Hello', string2='world')
        )

    def test_args_and_kwargs(self):
        """Both args and kwargs"""

        assert_equal(
            {"jsonrpc": "2.0", "method": "set_age", "params": [40, {"name": "Beau Barker"}]},
            rpc.request('set_age', 40, name='Beau Barker')
        )

    def test_multiple_args_and_kwargs(self):
        """Both args and kwargs"""

        assert_equal(
            {"jsonrpc": "2.0", "method": "set_age", "params": [39, 40, {"first": "Beau", "last": "Barker"}]},
            rpc.request('set_age', 39, 40, first='Beau', last='Barker')
        )

# Non-notifications (require a response)

    def test_method_only_request(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "get", "id": 1},
            rpc.request('get', response=True)
        )

    def test_one_arg_request(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "sqrt", "params": [1], "id": 1},
            rpc.request('sqrt', 1, response=True)
        )

    def test_two_args_request(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "add", "params": [1, 2], "id": 1},
            rpc.request('add', 1, 2, response=True)
        )

    def test_one_kwarg_request(self):
        assert_equal(
            {"jsonrpc": "2.0", "method": "find", "params": {"name": "Beau"}, "id": 1},
            rpc.request('find', name='Beau', response=True)
        )

    def test_two_kwargs_request(self):
        # Note in the resulting json, the params are in reverse order to how the
        # kwargs are called, because the args are sorted by key in the request
        # function. This is the only way to know exactly how the request will look
        # because the kwargs come in a random order. There is no way of getting the
        # actual order of the kwargs
        assert_equal(
            {"jsonrpc": "2.0", "method": "find", "params": {"age": 38, "name": "Beau"}, "id": 1},
            rpc.request('find', name='Beau', age=38, response=True)
        )

    def test_args_and_kwargs_request(self):
        # Includes both args and kwargs
        assert_equal(
            {"jsonrpc": "2.0", "method": "find", "params": {"age": 38, "name": "Beau"}, "id": 1},
            rpc.request('find', name='Beau', age=38, response=True)
        )

    def test_multiple_args_and_kwargs_request(self):
        # Both args and kwargs
        assert_equal(
            {"jsonrpc": "2.0", "method": "set_age", "params": [39, 40, {"first": "Beau", "last": "Barker"}], "id": 1},
            rpc.request('set_age', 39, 40, first='Beau', last='Barker', response=True)
        )



"""proxy.py"""

import json
import logging
import pkgutil

import requests
import jsonschema

from . import rpc
from . import exceptions

class Proxy:
    """This class acts as the remote server"""

    def __init__(self, endpoint):
        self.endpoint = endpoint

    def __getattr__(self, name):
        """Catch undefined methods and handle them as RPC requests.

        The technique is here:
        http://code.activestate.com/recipes/307618/
        """

        def attr_handler(*args, **kwargs):
            """Call self.request from here"""
            return self.request(name, *args, **kwargs)

        return attr_handler

    def request(self, method_name, *args, **kwargs):
        """Send the request and expect a response."""

        kwargs['response'] = True
        return self.handle_response(
            self.send_message(rpc.request(method_name, *args, **kwargs)), True)

    def notify(self, method_name, *args, **kwargs):
        """Really just an alias for _send()"""

        return self.handle_response(
            self.send_message(rpc.request(method_name, *args, **kwargs)), False)

    def send_message(self, request_dict):
        """Send the RPC request to the proxy server.

        Calls a procedure on another server.
        Raises RPCClientException: On any error caught.
        """

        # Log the request
        logging.info('--> '+json.dumps(request_dict))

        try:
            # Send the message
            r = requests.post(
                self.endpoint,
                headers={
                    'Content-Type': 'application/json; charset=utf-8'
                },
                json=request_dict
            )

            logging.info('<-- '+r.text.strip("\n"))

        except (requests.exceptions.InvalidSchema,
                requests.exceptions.RequestException):
            raise exceptions.ConnectionError()

        # Raise exception the HTTP status code was not 200, and there was no
        # response body, because this should be handled.
        if not len(r.text) and r.status_code != 200:
            raise exceptions.StatusCodeError(r.status_code)

        return r.text

    @staticmethod
    def handle_response(response_str, expected_response=True):
        """Processes the response from a request"""

        # A response was expected, but none was given?
        if expected_response and not len(response_str):
            raise exceptions.ReceivedNoResponse()

        # Was response given?
        if len(response_str):

            # Attempt to parse the response
            try:
                response_dict = json.loads(response_str)

            except ValueError:
                raise exceptions.ParseError()

            # A response was *not* expected, but one was given? It may not
            # be necessary to raise here. If we receive a response anyway,
            # can't we just ignore it?
            if not expected_response and 'result' in response_dict:
                raise exceptions.InvalidResponse()

            # Validate the response against the Response schema
            try:
                jsonschema.validate(
                    response_dict,
                    json.loads(pkgutil.get_data(
                        __name__, 'response-schema.json').decode('utf-8')))

            except jsonschema.ValidationError:
                raise exceptions.InvalidResponse()

            # If the response was "error", raise it, to ensure it's handled
            if 'error' in response_dict:
                raise exceptions.ReceivedErrorResponse(
                    response_dict['error']['code'],
                    response_dict['error']['message'])

            # Otherwise, surely we have a result to return
            print(response_dict['result'])

        return None



#!/usr/bin/env python
# pylint: disable=line-too-long
"""setup.py"""

from distutils.core import setup

setup(
    name='jsonrpcclient',
    packages=['jsonrpcclient'],
    package_data={'jsonrpcclient': ['response-schema.json']},
    install_requires=['requests', 'jsonschema'],
    version='1.0.3',
    description='JSON-RPC 2.0 client library for Python 3',
    author='Beau Barker',
    author_email='beauinmelbourne@gmail.com',
    url='https://bitbucket.org/beau-barker/jsonrpcclient',
    keywords=['json-rpc', 'json', 'api'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )
